"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("@typescript-eslint/types");
const constants_1 = require("../shared/constants");
const util_1 = require("../util");
const topLevelVariablesSet = new Set();
const EXPORTS = 'exports';
exports.default = (0, util_1.createRule)({
    defaultOptions: [],
    name: 'no-commonjs-exports',
    meta: {
        type: 'suggestion',
        docs: {
            description: 'Disable the use of CommonJS "exports" for exporting modules in your project.',
            recommended: false,
        },
        fixable: 'code',
        messages: {
            noCommonJSExports: (0, util_1.ruleMessageTemplate)({
                why: 'TypeScript official module system is ESM (ECMAScript Modules), Also the future of JavaScript in general.',
                linterMessage: `Calls to CommonJS "${EXPORTS}" should be replaced with ESM "export" syntax.`,
            }),
        },
        schema: [],
    },
    create(context) {
        return {
            Program(_node) {
                topLevelVariablesSet.clear();
            },
            MemberExpression(node) {
                if (!isModuleScope(context.getScope())) {
                    return;
                }
                if (isMemberExpressionAnExports(node)) {
                    const nodeLeftPropertyIdentifier = node.property;
                    context.report({
                        node,
                        messageId: 'noCommonJSExports',
                        fix(fixer) {
                            return [
                                ... // and only if no variable exists
                                (topLevelVariablesSet.has(nodeLeftPropertyIdentifier.name)
                                    ? []
                                    : [
                                        fixer.replaceTextRange([
                                            // from
                                            node.range[0],
                                            // to
                                            node.property.range[0],
                                        ], 'export const '),
                                    ]),
                            ];
                        },
                    });
                }
            },
            Identifier(node) {
                if (!isModuleScope(context.getScope())) {
                    return;
                }
                if (doesIdentifierIsExports(node)) {
                    context.report({
                        node,
                        messageId: 'noCommonJSExports',
                    });
                }
            },
            // Collect all top-level variables to check if automatic fix is possible
            VariableDeclaration(node) {
                if (!isModuleScope(context.getScope())) {
                    return;
                }
                for (const d of node.declarations) {
                    if (d.id.type === types_1.AST_NODE_TYPES.Identifier) {
                        topLevelVariablesSet.add(d.id.name);
                    }
                }
            },
        };
    },
});
function isModuleScope(scope) {
    return scope.variableScope.type === constants_1.CEScopeType.Module;
}
/** Detects cases like: `exports = {}` */
function doesIdentifierIsExports(node) {
    var _a, _b, _c;
    return Boolean(node.parent &&
        node.parent.type === types_1.AST_NODE_TYPES.AssignmentExpression &&
        node.parent.operator === '=' &&
        ((_a = node.parent.parent) === null || _a === void 0 ? void 0 : _a.type) === types_1.AST_NODE_TYPES.ExpressionStatement &&
        ((_b = node.parent.parent.parent) === null || _b === void 0 ? void 0 : _b.type) === types_1.AST_NODE_TYPES.Program &&
        ((_c = node.parent.parent.parent) === null || _c === void 0 ? void 0 : _c.sourceType) === constants_1.CEScopeType.Module);
}
/** Detects cases like: `exports.a = {}` */
function isMemberExpressionAnExports(node) {
    return (node.type === types_1.AST_NODE_TYPES.MemberExpression &&
        node.object.type === types_1.AST_NODE_TYPES.Identifier &&
        node.object.name === EXPORTS &&
        node.property &&
        node.property.type === types_1.AST_NODE_TYPES.Identifier);
}
//# sourceMappingURL=no-commonjs-exports.js.map